import MadX as _MadX

def madx2gmad(iFilename, oFilename ) :
    o = open(oFilename,'w')
    
    t = _MadX.Twiss(iFilename)
    
    s = t.data['S'] 

    
    kw  = t.data['KEYWORD'] 
    na  = t.data['NAME'] 
    l   = t.data['L']
    k0l = t.data['K0L']
    k1l = t.data['K1L']
    k2l = t.data['K2L'] 

    i = 0 
    
    stw = str() 
    for v in s : 
        if kw[i] == 'MULTIPOLE' :
            if k0l[i] != 0 : 
                stw = na[i]+': dipole, l='+str(l[i])
            if k1l[i] != 0 : 
                stw = na[i]+': quadrupole, l='+str(l[i])+' k1='+str(k1l[i])
            if k2l[i] != 0 :
                stw = na[i]+': sextupole, l='+str(l[i])+' k1='+str(k1l[i])
        elif kw[i] == 'DRIFT' : 
            stw = na[i]+': drift, l='+str(l[i])

        stw += ';\n'
        
        o.write(stw)

        i += 1



import os
import numpy as np
import pymadx


abs_difference = 1e-6

def within_tolerance(val, ref, tolerance=abs_difference):
    v = np.abs(val - ref) <= tolerance
    return np.all(v)


def _fn(filename):
    return os.path.join(os.path.dirname(__file__), "test_input", filename)


def test_data_tfs_load():
    d = pymadx.Data.Tfs(_fn("h6-simple.tfs"))
    assert len(d) == 226


def test_data_tfs_survey_load():
    d = pymadx.Data.Tfs(_fn("h6-survey.tfs"))
    assert len(d) == 226


def test_data_tfs_load_gz():
    d = pymadx.Data.Tfs(_fn("h6-positive-120gev-fm.tfs.gz"))
    assert len(d) == 226


def test_roto_translation():
    r = pymadx.Data.RotoTranslation2D(np.pi/6, np.array([3, 4]))
    w, l = 3, 5
    x0, y0 = 10, 1
    points = np.array([[x0, y0+0.5*w], [x0, y0-0.5*w], [x0-l, y0-0.5*w], [x0-l, y0+0.5*w]])
    rpoints = r * points
    ref = np.array([[10.41025404, 11.16506351],
                    [11.91025404,  8.5669873],
                    [7.58012702,   6.0669873],
                    [6.08012702,   8.66506351]])
    assert within_tolerance(rpoints, ref)


def test_roto_translation_rot_origin():
    x0, y0 = 10, 1
    r = pymadx.Data.RotoTranslation2D(np.pi/6, np.array([3, 4]), rotationOrigin=np.array([x0, y0]))
    w, l = 3, 5
    points = np.array([[x0, y0+0.5*w], [x0, y0-0.5*w], [x0-l, y0-0.5*w], [x0-l, y0+0.5*w]])
    rpoints = r * points
    ref = np.array([[12.25,        6.29903811],
                    [13.75,        3.70096189],
                    [9.41987298,   1.20096189],
                    [7.91987298,   3.79903811]])
    assert within_tolerance(rpoints, ref)

def test_roto_tranlsate_from_survey():
    d = pymadx.Data.Tfs(_fn("h6-survey.tfs"))
    r = d.GetRotoTranslationFromElementZX("MBXHC.X0410132")
    localPos = np.array([[0, 0]])
    globalPos = r * localPos
    ref = np.array([d['MBXHC.X0410132']['Z'], d['MBXHC.X0410132']['X']])
    assert within_tolerance(globalPos, ref)


def test_roto_tranlsate_from_survey_inverse():
    d = pymadx.Data.Tfs(_fn("h6-survey.tfs"))
    r = d.GetRotoTranslationFromElementZX("MBXHC.X0410132")
    rinv = r.Inverse()
    ele = d['MBXHC.X0410132']
    globalPos = np.array([[ele['Z'], ele['X']]])
    localPos = rinv * globalPos
    ref = np.array([[0, 0]])
    assert within_tolerance(localPos, ref)

test_roto_translation()
test_roto_translation_rot_origin()
test_roto_tranlsate_from_survey()
test_roto_tranlsate_from_survey_inverse()


#wrapper for clean interface
#wrapper functions should be placed in here and long complicated 

def Mad8ToMadx(inputfilename):
    """
    Convert a Mad8 file to the syntax of a MadX file.

    The output file name will be the input file name with
    the file extension replaced by .xsifx
    """
    _Mad8.Mad8ToMadX(inputfilename)



import tarfile 
from Tfs import Cast 
import numpy as _numpy
import copy as _copy

class TfsArray : 

    def __init__(self,filename=None) : 
        self.header  = {}
        self.columns = [] 
        self.formats = [] 
        self.segment = [] 
        self.data    = []
        self.nitems  = 0
        self.filename= filename
        if self.filename != None : 
            self.Load(self.filename) 


    def Clear(self) :
        """ 
        Clear() 
        empty all internal variables 
        """
        self.__init__()

    def Load(self,filename) :
        """
        Load('filename.tfs')        
        read the tfs file and prepare data structures
        """

        # open gz file 
        if 'gz' in filename :
            print 'pymadx.TfsArray.Load> zipped file'
            tar = tarfile.open(filename,'r')
            f = tar.extractfile(tar.firstmember)
        else:
            print 'pymadx.TfsArray> normal file'
            f = open(filename)


        self.isegment = -1
        self.columns.append('SEGMENT')
        self.formats.append('%d')

        for line in f :
            sl = line.strip('\n').split()
            if line[0] == "@" : 
                # header
                # print 'pymadx.TfsArray> header',line 
                self.header[sl[1]] = sl[-1].replace('"','')
            elif line[0] == '*' : 
                # column name 
                # print 'pymadx.TfsArray> columns',line 
                self.columns.extend(sl[1:]) #miss *                
            elif line[0] == '$' : 
                # column type
                # print 'pymadx.TfsArray> type',line 
                self.formats.extend(sl[1:]) #miss $              
            elif sl[0] == '#segment' : 
                # entry marker 
                # print 'pymadx.TfsArray> marker',line, self.nitems,isegment
                self.isegment=self.isegment+1
                
            else :                 
                # data
                d = [Cast(item) for item in sl]
                d.insert(0,self.isegment) 
                self.data.append(d)    # put in data dict by name
                self.nitems += 1 # keep tally of number of items                

            
        # make numpy array for convenience
        self.dataArray = _numpy.array(self.data)

    def __repr__(self):
        s =  ''
        s += 'pymadx.TfsArray instance\n'
        s += str(self.nitems) + ' items in lattice\n'
        return s

    def __iter__(self):
        self._iterindex = -1
        return self

    def next(self):
        if self._iterindex == len(self.sequence)-1:
            raise StopIteration
        self._iterindex += 1
        return self.GetRow(self._iterindex)

    def __getitem__(self,index):
        #return single item or slice of lattice
        if type(index) == slice:
            #prepare step integer - allow reverse stepping too
            if index.stop > index.start:
                index = slice(index.start,index.stop,1)
            else:
                index = slice(index.start,index.stop,-1)
            return [self.GetRow(i) for i in range(index.start,index.stop,index.step)]
        elif type(index) == int:
            return self.GetRow(index)
        else:
            raise ValueError("argument not an index or a slice")

    def ColumnIndex(self,columnstring):
        """
        ColumnIndex(columnname):
        
        return the index to the column matching the name
        
        REMEMBER: excludes the first column NAME
        0 counting

        """
        return self.columns.index(columnstring)

    def NameFromIndex(self,index):
        return self.dataArray[index,self.columns.index['NAME']]
            
    def GetColumn(self, colname) :
        colindex = self.columns.index(colname) 
        return self.dataArray[:,colindex]

    def GetRow(self, rowindex) : 
        return self.dataArray[rowindex,:]

    def GetSegment(self, segmentindex) : 
        segcut = self.dataArray[:,0] == segmentindex
        c = _copy.deepcopy(self)
        c.data      = self.data
        c.dataArray = self.dataArray[segcut,:]
        c.nitems    = len(c.dataArray)        
        return c

    def InterrogateItem(self,index):
        """
        InterrogateItem(index)
        
        Print out all the parameters and their
        names for a particlular element in the 
        sequence identified by name
        """
        for i,parameter in enumerate(self.columns):
            print parameter.ljust(10,'.'),self.dataArray[index,i]

    def GetElementsOfType(self,typename) :
        """
        GetElementsOfType(typename) 
        
        Returns a list of the names of elements of a certain type

        typename can be a single string or a tuple or list of strings

        ie 
        GetElementsOfType('SBEND')
        GetElementsOfType(['SBEND','RBEND'])
        GetElementsOfType(('SBEND','RBEND','QUADRUPOLE'))

        """        
        i = self.ColumnIndex('KEYWORD')
        types = self.GetColumn('KEYWORD')
        return [t for t in types if t in typename]

    def ReportPopulations(self):
        """
        Print out all the population of each type of
        element in the beam line (sequence)
        """
        print 'Filename > ',self.filename
        print 'Total number of items > ',self.nitems
        i = self.ColumnIndex('KEYWORD')
        types = set(self.GetColumn('KEYWORD'))
        populations = [(len(self.GetElementsOfType(key)),key) for key in types]
        print 'Type'.ljust(15,'.'),'Population'
        for item in sorted(populations)[::-1]:
            print item[1].ljust(15,'.'),item[0]



import os
import pymadx
import pytest

_skip_cpymad_tests = False
try:
    import cpymad
    import cpymad.madx
except ImportError:
    _skip_cpymad_tests = True

if _skip_cpymad_tests:
    pytestmark = pytest.mark.xfail(run=True, reason="requires cpymad")


def _fn(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def test_Tfs_from_cpymad_table():
    m = cpymad.madx.Madx(stdout=None)
    m.call(file=_fn("test_input/h6.madx"))
    tt = m.table['twiss']
    tfs = pymadx.Data.Tfs(tt)
    print(tfs)



from .. import Data as _Data
from .. import Plot as _Plot

from os import path as _path

import matplotlib.pyplot as _plt

def _LoadTfsInfput(tfs_in, name):
    """
    Return tfs_in as a Tfs instance, which should either be a path
    to a TFS file or a Tfs instance, and in either case, generate a
    name if None is provided, and return that as well.
    """
    if isinstance(tfs_in, str):
        if not _path.isfile(tfs_in):
            raise IOError("file \"{}\" not found!".format(tfs_in))
        name = (_path.splitext(_path.basename(tfs_in))[0]
                if name is None else name)
        return _Data.Tfs(tfs_in), name
    try:
        name = tfs_in.filename if name is None else name
        return tfs_in, name
    except AttributeError:
        raise TypeError("Expected Tfs input is neither a "
                        "file path nor a Tfs instance: {}".format(tfs_in))


# use closure to avoid tonnes of boilerplate code as happened with
# MadxBdsimComparison.py
def _MakePlotter(plot_info_tuples, x_label, y_label, title, factor=1.0):
    def f_out(first, second, first_name=None, second_name=None, **kwargs):
        """first and second should be tfs files."""
        first, first_name = _LoadTfsInfput(first, first_name)
        second, second_name = _LoadTfsInfput(second, second_name)

        plot = _plt.figure(title, figsize=(9,5), **kwargs)
        # Loop over the variables in plot_info_tuples and draw the plots.
        colours = ["tab:blue", "tab:orange"]
        for colour, (var, legend_name) in zip(colours, plot_info_tuples):
            _plt.plot(first.GetColumn('S'), first.GetColumn(var)*factor, c=colour,
                      label="{}: {}".format(legend_name, first_name), **kwargs)
            _plt.plot(second.GetColumn('S'), second.GetColumn(var)*factor, '--', c=colour,
                      label="{}: {}".format(legend_name, second_name), **kwargs)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        _Plot.AddMachineLatticeToFigure(plot, first)
        _plt.show(block=False)
        return plot
    return f_out


def _MakePlotterWithScale(plot_info_tuples, x_label, y_label, title, zeroLine=False):
    def f_out(first, second, first_name=None, second_name=None, dSSecond=0, **kwargs):
        """first and second should be tfs files."""
        first, first_name = _LoadTfsInfput(first, first_name)
        second, second_name = _LoadTfsInfput(second, second_name)
        cmap = _plt.get_cmap('tab20')
        plot = _plt.figure(title, figsize=(9,5), **kwargs)
        # Loop over the variables in plot_info_tuples and draw the plots.
        for i, (var, scale, legend_name) in enumerate(plot_info_tuples):
            _plt.plot(first.GetColumn('S'),
                      first.GetColumn(var)*scale,
                      c=cmap(2*i),
                      label="{}: {}".format(legend_name, first_name),
                      **kwargs)
            _plt.plot(second.GetColumn('S')+dSSecond,
                      second.GetColumn(var)*scale,
                      c=cmap(2*i + 1), ls='--',
                      label="{}: {}".format(legend_name, second_name),
                      **kwargs)
        if zeroLine:
            s = first.GetColumn('S')
            _plt.plot([s[0], s[-1]], [0, 0], c='grey', alpha=0.3)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        _Plot.AddMachineLatticeToFigure(plot, first)
        _plt.show(block=False)
        return plot
    return f_out


"""
Functions for comparing the optical distributions of a madx model
and a Transport model.

Functions for plotting the individual optical functions, and an
eighth, helper function ''compare_all_optics``, to plot display all
seven in one go.
"""
import datetime as _datetime
import matplotlib.pyplot as _plt
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages

from ._CompareCommon import _LoadTfsInfput
from .. import Plot as _Plot

from pytransport import Reader as _Reader

_BETA = [("BETX", "Beta_x", r'$\beta_{x}$'),
         ("BETY", "Beta_y", r'$\beta_{y}$')]

_ALPHA = [("ALFX", "Alpha_x", r"$\alpha_{x}$"),
          ("ALFY", "Alpha_y", r"$\alpha_{y}$")]

_DISP = [("DXBETA", "Disp_x", r"$D_{x}$ / $\beta$"),
         ("DYBETA", "Disp_y", r"$D_{y}$ / $\beta$")]

_DISP_P = [("DPXBETA", "Disp_xp", r"$D_{p_{x}}$"),
           ("DPYBETA", "Disp_yp", r"$D_{p_{y}}$")]

_SIGMA = [("SIGMAX", "Sigma_x", r"$\sigma_{x}$"),
          ("SIGMAY", "Sigma_y", r"$\sigma_{y}$")]

_SIGMA_P = [("SIGMAXP", "Sigma_xp", r"$\sigma_{xp}$"),
            ("SIGMAYP", "Sigma_yp", r"$\sigma_{yp}$")]

# use closure to avoid tonnes of boilerplate code as happened with
# MadxBdsimComparison.py
def _make_plotter(plot_info_tuples, x_label, y_label, title):
    def f_out(first, second, first_name=None, second_name=None, **kwargs):
        """first and second should be tfs files."""
        first, first_name = _LoadTfsInfput(first, first_name)
        second = _Reader.GetOptics(second)

        plot = _plt.figure(title, figsize=(9,5), **kwargs)
        # Loop over the variables in plot_info_tuples and draw the plots.
        for madxVar, transVar, legend_name in plot_info_tuples:
            _plt.plot(first.GetColumn('S'), first.GetColumn(madxVar),
                      label="{}: {}".format(legend_name, first_name), **kwargs)
            _plt.plot(second.GetColumn('S'), second.GetColumn(transVar),
                      label="{}: {}".format(legend_name, second_name), **kwargs)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        _Plot.AddMachineLatticeToFigure(plot, first)
        _plt.show(block=False)
        return plot
    return f_out

PlotBeta   = _make_plotter(_BETA,    "S / m", r"$\beta_{x,y}$ / m",      "Beta")
PlotAlpha  = _make_plotter(_ALPHA,   "S / m", r"$\alpha_{x,y}$ / m",     "Alpha")
PlotDisp   = _make_plotter(_DISP,    "S / m", r"$D_{x,y} / m$",          "Dispersion")
PlotDispP  = _make_plotter(_DISP_P,  "S / m", r"$D_{p_{x},p_{y}}$ / m",  "Momentum_Dispersion")
PlotSigma  = _make_plotter(_SIGMA,   "S / m", r"$\sigma_{x,y}$ / m",     "Sigma")
PlotSigmaP = _make_plotter(_SIGMA_P, "S / m", r"$\sigma_{xp,yp}$ / rad", "SigmaP")
#PlotMean   = _make_plotter(_MEAN,    "S / m", r"$\bar{x}, \bar{y}$ / m", "Mean")


def MADXVsTRANSPORT(first, second, first_name=None,
               second_name=None, saveAll=True, 
               outputFileName=None, **kwargs):
    """
    Display all the optical function plots for the two input optics files.
    """
    figures = [
    PlotBeta(first, second, first_name=first_name,
             second_name=second_name, **kwargs),
    PlotAlpha(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotDisp(first, second, first_name=first_name,
             second_name=second_name, **kwargs),
    PlotDispP(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotSigma(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotSigmaP(first, second, first_name=first_name,
               second_name=second_name, **kwargs)
    ]

    if saveAll:
        if outputFileName is not None:
            output_filename = outputFileName
        else:
            output_filename = "optics-report.pdf"
        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} VS {} Optical Comparison".format(first_name, second_name)
            d['CreationDate'] = _datetime.datetime.today()
        print("Written ", output_filename)



# Wrapper for pytransport conversion.

from pymadx import Builder as _mdBuilder


from pytransport.Convert import _Convert
from pytransport.Data import ConversionData


def Transport2Madx(inputfile,
                   particle='proton',
                   distrType='gauss',
                   outputDir='madx',
                   debug=False,
                   dontSplit=False,
                   keepName=False,
                   combineDrifts=False):
    """
    **Transport2Madx** convert a Transport input or output file into a madx input file for madx

    +-------------------------------+-------------------------------------------------------------------+
    | **inputfile**                 | dtype = string                                                    |
    |                               | path to the input file                                            |
    +-------------------------------+-------------------------------------------------------------------+
    | **particle**                  | dtype = string. Optional, default = "proton"                      |
    |                               | the particle species                                              |
    +-------------------------------+-------------------------------------------------------------------+
    | **distrType**                 | dtype = string. Optional, Default = "gauss".                      |
    |                               | the beam distribution type. Can be either gauss or gausstwiss.    |
    +-------------------------------+-------------------------------------------------------------------+
    | **outputDir**                 | dtype=string. Optional, default = "gmad"                          |
    |                               | the output directory where the files will be written              |
    +-------------------------------+-------------------------------------------------------------------+
    | **debug**                     | dtype = bool. Optional, default = False                           |
    |                               | output a log file (inputfile_conversion.log) detailing the        |
    |                               | conversion process, element by element                            |
    +-------------------------------+-------------------------------------------------------------------+
    | **dontSplit**                 | dtype = bool. Optional, default = False                           |
    |                               | the converter splits the machine into multiple parts when a beam  |
    |                               | is redefined in a Transport lattice. dontSplit overrides this and |
    |                               | forces the machine to be written to a single file                 |
    +-------------------------------+-------------------------------------------------------------------+
    | **keepName**                  | dtype = bool. Optional, default = False                           |
    |                               | keep the names of elements as defined in the Transport inputfile. |
    |                               | Appends element name with _N where N is an integer if the element |
    |                               | name has already been used                                        |
    +-------------------------------+-------------------------------------------------------------------+
    | **combineDrifts**             | dtype = bool. Optional, default = False                           |
    |                               | combine multiple consecutive drifts into a single drift           |
    +-------------------------------+-------------------------------------------------------------------+

    Example:

    >>> Transport2Madx(inputfile)

    Writes converted machine to disk. Reader automatically detects if the supplied input file is a Transport input
    file or Transport output file.

    """
    converter = _Convert(ConversionData(inputfile=inputfile, options=None, machine=_mdBuilder.Machine(),
                                        particle=particle, debug=debug, distrType=distrType, gmad=False, gmadDir='',
                                        madx=True, madxDir=outputDir, dontSplit=dontSplit, keepName=keepName,
                                        combineDrifts=combineDrifts))
    # automatically convert
    converter.Convert()

